# The MIT License (MIT)
# Copyright (c) 2025 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import aiohttp
import asyncio
import random
import time
from typing import Optional

from .constants import DEFAULT_NUM_RETRIES
from .constants import DEFAULT_RETRY_BACKOFF_MAX
from .constants import DEFAULT_RETRY_BACKOFF_BASE
from .constants import LOG

async def _run_with_session_executor(async_function, *params, headers):
    async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit=50),
            headers=headers,
            trust_env=True
    ) as session:
        return await async_function(session, *params)


class SessionExecutor:

    def __init__(
            self,
            user_agent: str = None,
            enable_warnings: bool = False,
            num_retries: int = DEFAULT_NUM_RETRIES,
            retry_backoff_max: int = DEFAULT_RETRY_BACKOFF_MAX,
            retry_backoff_base: float = DEFAULT_RETRY_BACKOFF_BASE,
    ):
        self._headers = {'User-Agent': user_agent} if user_agent else None
        self._enable_warnings = enable_warnings
        self._num_retries = num_retries
        self._retry_backoff_max = retry_backoff_max
        self._retry_backoff_base = retry_backoff_base

    def run_with_session(self, async_function, *params):
        # See https://github.com/aio-libs/aiohttp/blob/master/docs/
        # client_advanced.rst#graceful-shutdown
        loop = asyncio.new_event_loop()
        coro = _run_with_session_executor(
            async_function, *params, headers=self._headers
        )
        result = loop.run_until_complete(coro)
        # Short sleep to allow underlying connections to close
        loop.run_until_complete(asyncio.sleep(1.))
        loop.close()
        return result

    def get_response_content(self, url: str) -> Optional[bytes]:
        return self.run_with_session(self.get_response_content_from_session, url)

    async def get_response_content_from_session(
            self, session: aiohttp.ClientSession,
            url: str
    ) -> Optional[bytes]:
        num_retries = self._num_retries
        retry_backoff_max = self._retry_backoff_max  # ms
        retry_backoff_base = self._retry_backoff_base
        for i in range(num_retries):
            resp = await session.request(method='GET', url=url)
            if resp.status == 200:
                try:
                    content = await resp.read()
                    return content
                except aiohttp.client_exceptions.ClientPayloadError as cpe:
                    error_message =str(cpe)
            elif 500 <= resp.status < 600:
                if self._enable_warnings:
                    error_message = f'Error {resp.status}: Cannot access url.'
                    LOG.warning(error_message)
                return None
            elif resp.status == 429:
                error_message = "Error 429: Too Many Requests."
            else:
                break
            # Retry after 'Retry-After' with exponential backoff
            retry_min = int(resp.headers.get('Retry-After', '100'))
            retry_backoff = random.random() * retry_backoff_max
            retry_total = retry_min + retry_backoff
            if self._enable_warnings:
                retry_message = \
                    f'{error_message} ' \
                    f'Attempt {i + 1} of {num_retries} to retry after ' \
                    f'{"%.2f" % retry_min} + {"%.2f" % retry_backoff} = ' \
                    f'{"%.2f" % retry_total} ms ...'
                LOG.info(retry_message)
            time.sleep(retry_total / 1000.0)
            retry_backoff_max *= retry_backoff_base
        return None



import jsonschema
import unittest

from xcube.core.store import DatasetDescriptor
from xcube.util.jsonschema import JsonObjectSchema

from xcube_cci.dtaccess import DataTreeDescriptor

class DataTreeDescriptorTest(unittest.TestCase):
    def test_get_schema(self):
        schema = DataTreeDescriptor.get_schema()
        self.assertIsInstance(schema, JsonObjectSchema)

    def test_from_dict_no_data_id(self):
        descriptor_dict = dict()
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            DataTreeDescriptor.from_dict(descriptor_dict)

    def test_from_dict_wrong_data_type(self):
        descriptor_dict = dict(data_id="xyz", data_type="tsr")
        with self.assertRaises(ValueError) as cm:
            DataTreeDescriptor.from_dict(descriptor_dict)
        self.assertEqual("unknown data type 'tsr'", f"{cm.exception}")

    def test_from_dict_derived_type(self):
        descriptor_dict = dict(data_id="xyz", data_type="datatree")
        descriptor = DataTreeDescriptor.from_dict(descriptor_dict)
        self.assertIsNotNone(descriptor)
        self.assertEqual("xyz", descriptor.data_id)
        self.assertEqual("datatree", descriptor.data_type.alias)

    def test_from_dict_full(self):
        descriptor_dict = dict(
            data_id="xyz",
            data_type="datatree",
            crs="EPSG:9346",
            bbox=(10.0, 20.0, 30.0, 40.0),
            spatial_res=20.0,
            time_range=("2017-06-05", "2017-06-27"),
            time_period="daily",
            coords=dict(
                rtdt=dict(
                    name="rtdt",
                    dtype="rj",
                    dims=("rtdt",),
                    attrs=dict(ssd=6, zjgrhgu="hgtr"),
                )
            ),
            dims=dict(x=1, y=2, z=3),
            data_vars=dict(
                xf=dict(
                    name="xf",
                    dtype="rj",
                    dims=("dfjhrt", "sg"),
                    attrs=dict(ssd=4, zjgrhgu="dgfrf"),
                )
            ),
            dataset=dict(
                data_id="abc",
                data_type="dataset",
            ),
            data_nodes=dict(
                first_inner_tree=dict(
                    data_id="first_inner_tree",
                    data_type="datatree",
                ),
                second_inner_tree=dict(
                    data_id="second_inner_tree",
                    data_type="datatree",
                )
            ),
            attrs=dict(dzus=236, tgr7h="rt5", df="s8fd4w5"),
            open_params_schema=dict(
                type="object",
                properties=dict(
                    variable_names=dict(type="array", items=dict(type="string"))
                ),
            ),
        )
        descriptor = DataTreeDescriptor.from_dict(descriptor_dict)
        self.assertIsNotNone(descriptor)
        self.assertEqual("xyz", descriptor.data_id)
        self.assertEqual("datatree", descriptor.data_type.alias)
        self.assertEqual("EPSG:9346", descriptor.crs)
        self.assertEqual((10.0, 20.0, 30.0, 40.0), descriptor.bbox)
        self.assertEqual(20.0, descriptor.spatial_res)
        self.assertEqual(("2017-06-05", "2017-06-27"), descriptor.time_range)
        self.assertEqual("daily", descriptor.time_period)
        self.assertEqual(1, len(descriptor.coords))
        self.assertTrue("rtdt" in descriptor.coords)
        self.assertEqual(dict(x=1, y=2, z=3), descriptor.dims)
        self.assertEqual(1, len(descriptor.data_vars))
        self.assertTrue("xf" in descriptor.data_vars)
        self.assertTrue("abc" in descriptor.dataset.data_id)
        self.assertEqual("dataset", descriptor.dataset.data_type.alias)
        self.assertIn("first_inner_tree", list(descriptor.data_nodes.keys()))
        self.assertIn("first_inner_tree", descriptor.data_nodes.get("first_inner_tree").data_id)
        self.assertIn("datatree", descriptor.data_nodes.get("first_inner_tree").data_type.alias)
        self.assertIn("second_inner_tree", list(descriptor.data_nodes.keys()))
        self.assertIn("second_inner_tree", descriptor.data_nodes.get("second_inner_tree").data_id)
        self.assertIn("datatree", descriptor.data_nodes.get("second_inner_tree").data_type.alias)
        self.assertEqual(236, descriptor.attrs.get("dzus", None))
        self.assertEqual("rt5", descriptor.attrs.get("tgr7h", None))
        self.assertEqual("s8fd4w5", descriptor.attrs.get("df", None))
        self.assertEqual("object", descriptor.open_params_schema.get("type", None))

    def test_from_dict_datatree_descriptors_as_dict(self):
        descriptor_dict = dict(
            data_id="xyz",
            data_type="datatree",
            crs="EPSG:4326",
            data_nodes=dict(
                first_inner_tree=dict(
                    data_id="first_inner_tree",
                    data_type="datatree",
                )
            ),
        )
        descriptor = DataTreeDescriptor.from_dict(descriptor_dict)
        self.assertEqual("xyz", descriptor.data_id)
        self.assertEqual("datatree", descriptor.data_type.alias)
        self.assertEqual("EPSG:4326", descriptor.crs)
        self.assertEqual(1, len(descriptor.data_nodes))
        self.assertTrue("first_inner_tree" in descriptor.data_nodes)
        self.assertIsInstance(descriptor.data_nodes.get("firstInnerTree"), DataTreeDescriptor)

    def test_from_dict_datatree_descriptors_as_dict(self):
        descriptor_dict = dict(
            data_id="xyz",
            data_type="datatree",
            crs="EPSG:4326",
            dataset=dict(
                data_id="abc",
                data_type="dataset",
            ),
        )
        descriptor = DataTreeDescriptor.from_dict(descriptor_dict)
        self.assertEqual("xyz", descriptor.data_id)
        self.assertEqual("datatree", descriptor.data_type.alias)
        self.assertEqual("EPSG:4326", descriptor.crs)
        self.assertIsInstance(descriptor.dataset, DatasetDescriptor)

    def test_to_dict(self):
        # coords = dict(
        #     rtdt=VariableDescriptor(
        #         name="rtdt",
        #         dtype="rj",
        #         dims=("rtdt",),
        #         chunks=(2,),
        #         attrs=dict(ssd=6, zjgrhgu="hgtr"),
        #     )
        # )
        # var_descriptors = dict(
        #     xf=VariableDescriptor(
        #         name="xf",
        #         dtype="rj",
        #         dims=("dfjhrt", "sg"),
        #         chunks=(2, 3),
        #         attrs=dict(ssd=4, zjgrhgu="dgfrf"),
        #     )
        # )
        dataset_descriptor = DatasetDescriptor(
            data_id="xyz",
            crs="EPSG:9346",
            bbox=(10.0, 20.0, 30.0, 40.0),
            spatial_res=20.0,
            time_range=("2017-06-05", "2017-06-27"),
            time_period="daily",
            dims=dict(x=1, y=2, z=3),
            attrs=dict(dzus=236, tgr7h="rt5", df="s8fd4w5"),
        )
        inner_datatree_descriptor = DataTreeDescriptor(
            data_id="dti",
            crs="EPSG:9346",
        )
        datatree_descriptor = DataTreeDescriptor(
            data_id="dt",
            dataset=dataset_descriptor,
            data_nodes=dict(
                dti=inner_datatree_descriptor
            )
        )
        descriptor_dict = datatree_descriptor.to_dict()
        self.assertEqual(
            dict(
                data_id="dt",
                data_type="datatree",
                dataset=dict(
                    data_id="xyz",
                    data_type="dataset",
                    crs="EPSG:9346",
                    bbox=[10.0, 20.0, 30.0, 40.0],
                    spatial_res=20.0,
                    time_range=("2017-06-05", "2017-06-27"),
                    time_period="daily",
                    dims=dict(x=1, y=2, z=3),
                    attrs=dict(dzus=236, tgr7h="rt5", df="s8fd4w5")
                ),
                data_nodes=dict(
                    dti=dict(
                        data_id="dti",
                        data_type="datatree",
                        crs="EPSG:9346"
                    )
                )

            ),
            descriptor_dict,
        )



# The MIT License (MIT)
# Copyright (c) 2025 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Hashable, Mapping, MutableMapping
import dask
import numpy as np
from typing import Any, List, Optional
import warnings
import xarray as xr

from xcube.core.normalize import normalize_dataset
from xcube.core.store import DataType, DataTypeLike
from xcube.core.store import DataDescriptor, DatasetDescriptor, VariableDescriptor
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonIntegerSchema, JsonNumberSchema, JsonObjectSchema

from .cciodp import CciOdp
from .chunkstore import CciChunkStore

DATATREE_TYPE = DataType(xr.DataTree, ["datatree", "xarray.DataTree"])
DataType.register_data_type(DATATREE_TYPE)

class DataTreeDescriptor(DataDescriptor):
    """A descriptor for a tree-like hierarchical collection of gridded,
    N-dimensional datasets.

    The datasets are represented by xarray.Datasets, the collections by
    xarray.DataTrees.

    Args:
        data_id: An identifier for the data tree
        data_type: The data type of the data described
        crs: A coordinate reference system identifier, as an EPSG, PROJ
            or WKT string
        bbox: A bounding box of the data
        time_range: Start and end time delimiting this data's temporal
            extent
        time_period: The data's periodicity if it is evenly temporally
            resolved
        spatial_res: The spatial extent of a pixel in crs units
        dims: A mapping of the dataset's dimensions to their sizes
        coords: mapping of the dataset's data coordinate names to
            instances of :class:`VariableDescriptor`
        data_vars: A mapping of the dataset's variable names to
            instances of :class:`VariableDescriptor`
        dataset: A dataset associated with this node
        data_nodes: A mapping of data identifiers to DataTreeDescriptors
        attrs: A mapping containing arbitrary attributes of the dataset
        open_params_schema: A JSON schema describing the parameters that
            may be used to open this data
    """

    def __init__(
        self,
        data_id: str,
        *,
        data_type: DataTypeLike = DATATREE_TYPE,
        crs: str = None,
        bbox: tuple[float, float, float, float] = None,
        time_range: tuple[Optional[str], Optional[str]] = None,
        time_period: str = None,
        spatial_res: float = None,
        dims: Mapping[str, int] = None,
        coords: Mapping[str, "VariableDescriptor"] = None,
        data_vars: Mapping[str, "VariableDescriptor"] = None,
        dataset: DatasetDescriptor = None,
        data_nodes: Mapping[str, "DataTreeDescriptor"] = None,
        attrs: Mapping[Hashable, any] = None,
        open_params_schema: JsonObjectSchema = None,
        **additional_properties,
    ):
        super().__init__(
            data_id=data_id,
            data_type=data_type,
            crs=crs,
            bbox=bbox,
            time_range=time_range,
            time_period=time_period,
            open_params_schema=open_params_schema,
        )
        assert_true(
            DATATREE_TYPE.is_super_type_of(data_type),
            f"illegal data_type,"
            f" must be compatible with {DATATREE_TYPE!r}",
        )
        if additional_properties:
            warnings.warn(
                f"Additional properties received;"
                f" will be ignored: {additional_properties}"
            )
        self.dims = dict(dims) if dims else None
        self.spatial_res = spatial_res
        self.coords = coords if coords else None
        self.data_vars = data_vars if data_vars else None
        self.attrs = _attrs_to_json(attrs) if attrs else None
        self.dataset = dataset if dataset else None
        self.data_nodes = data_nodes if data_nodes else None

    @classmethod
    def get_schema(cls, recursion_depth: int = 0) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            dims=JsonObjectSchema(additional_properties=JsonIntegerSchema(minimum=0)),
            spatial_res=JsonNumberSchema(exclusive_minimum=0.0),
            coords=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            data_vars=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            dataset=DatasetDescriptor.get_schema(),
            attrs=JsonObjectSchema(additional_properties=True),
        )

        if recursion_depth < 10:
            schema.properties.update(
                data_nodes=JsonObjectSchema(
                    additional_properties=DataTreeDescriptor.get_schema(recursion_depth=recursion_depth+1)
                ),
            )
        schema.required = ["data_id", "data_type"]
        schema.additional_properties = False
        schema.factory = cls
        return schema


def _attrs_to_json(attrs: Mapping[Hashable, Any]) -> Optional[dict[str, Any]]:
    new_attrs: dict[str, Any] = {}
    for k, v in attrs.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()
        elif isinstance(v, dask.array.Array):
            v = np.array(v).tolist()
        if isinstance(v, float) and np.isnan(v):
            v = None
        new_attrs[str(k)] = v
    return new_attrs


class LazyDataTree(xr.DataTree):
    def __init__(self, name=None, dataset=None, children=None):
        super().__init__(name=name, dataset=None, children={})
        self._children = children or {}


class DataTreeMapping(MutableMapping):
    def __init__(
            self,
            base_id: str,
            specifiers: List[str],
            cci_odp: CciOdp,
            normalize_data: True,
            var_names: List[str] = None,
            pattern: str = None,
            **cci_kwargs
    ):
        self._base_id = base_id
        self._specifiers = specifiers
        self._cci_odp = cci_odp
        self._normalize_data = normalize_data
        self._var_names = var_names
        self._pattern = pattern
        self._cci_kwargs = cci_kwargs
        self._loaded = {}

    def __getitem__(self, key):
        if key not in self._loaded:
            if key not in self._specifiers:
                raise ValueError(f"Unsupported key {key}")
            if self._var_names is None or self._pattern is None:
                data_id = f"{self._base_id}~{key}"
                chunk_store = CciChunkStore(self._cci_odp, data_id, self._cci_kwargs)
                ds = xr.open_zarr(chunk_store, consolidated=False)
                ds.zarr_store.set(chunk_store)
            else:
                var_datasets = []
                for var_name in self._var_names:
                    var_key = self._pattern.format(var_name=var_name, specifier=key)
                    data_id = f"{self._base_id}~{var_key}"
                    chunk_store = CciChunkStore(self._cci_odp, data_id, self._cci_kwargs)
                    ds = xr.open_zarr(chunk_store, consolidated=False)
                    ds.zarr_store.set(chunk_store)
                    var_datasets.append(ds)
                ds = xr.merge(var_datasets)
            if self._normalize_data:
                ds = normalize_dataset(ds)
            self._loaded[key] = xr.DataTree(name=key, dataset=ds)
        return self._loaded[key]

    def __setitem__(self, key, value):
        if isinstance(value, xr.DataTree):
            self._loaded[key] = value
        else:
            raise ValueError("Value must be DataTree")

    def __delitem__(self, key):
        self._specifiers.pop(key)
        self._loaded.pop(key, None)

    def __iter__(self):
        return iter(self._specifiers)

    def __len__(self):
        return len(self._specifiers)

    def __repr__(self):
        return f"<LazyMapping: keys={self._specifiers} loaded={list(self._loaded)}>"
